#include <vector>
#include <allegro.h>
#include <math.h>
#include <stdlib.h>
#include <iostream>
#include <ostream>

using namespace std;

#include "common.h"
#include "classes.h"
#include "data.h"
#include "manager.h"
#include "shot.h"
#include "playerinfo.h"

// Radians per second.
static const float TurnSpeed = (90.0 / 180.0) * M_PI;
static const float MartianGravity = 4; // ms^-2 (metres per second per second)
static const float PixelsPerMetre = 10;
static const float DebrisSpeed = 100; // Pixels per second?
static const float ShieldRechargeTime = 90; // Seconds, full recharge
static const float FireDelay = 1; // seconds, between firing.
static const float ShotSpeed = 120; // Pixels per second?
static const float CollisionDamage = 100;
static const float TransferRate = 1.0/5.0; // Amount of tank, per second.
static const float MaxPickupSpeed = 100;

static void ApplyGravity(GameObject *obj, float quantum)
{
	obj->velocity.y += MartianGravity * PixelsPerMetre * quantum;
}

GameObject::~GameObject()
{
	// DO Nothing
};

void GameObject::tick(float quantum)
{
	// Nothing
}

void GameObject::save(ostream &out)
{
	// Nothing.
}

Ship::Ship() // Default constructor.
{
	velocity.x = 0;
	velocity.y = 0;
	angle = 0;
	dead = false; // important!
	radius = 23;
	mass = 10000; // Kgs
	// Needs to be at least mass * martiangravity otherwise player will be in trouble.
	sprite = (BITMAP *) datafile[SHIP_TGA].dat;
	debristime = 0;
	shieldstress = 0; // for the fadey thing
	shieldpower = 1.0; // show strong it is.
	ammotype = Hydrochloric;
	ammo = 1.0;
	oxygen = 1.0;
	hydrogen = 1.0;
	panelbackground = (BITMAP *) datafile[INFOPANEL_TGA].dat;
	panel = create_bitmap(panelbackground->w, panelbackground->h);
	updatepanel();
	// Fuel burn rate - depends on engine.
	thrust = 120000; // Newtons.
	burnrate = 1.0 / 20.0; // 20 seconds fuel.
	ammorate = 1.0 / 20.0;
	if (playerinfo.weapon == 2)
		ammorate = ammorate / 2;
	if (playerinfo.engine >= 2)
		burnrate = burnrate * (2.0 / 3.0);
	if (playerinfo.engine >= 3)
		thrust = thrust * 1.5;
	firetime = 0;
	connecteditem = 0;
	criticalhit = false;
	endlevel = 0;
}

void Ship::drawpanel()
{
	int x = screenoffsetx + 576;
	int y = screenoffsety;
	blit(panel, vscreen, 0,0,x,y, panel->w, panel->h);
	addrect(x,y,panel->w,panel->h);
}

static void drawslider(BITMAP *bmp, int x, int y, int colour, float amount)
{
	int width = 30, height = 6;
	int offset = (int) (amount * (float) width);
	rectfill(bmp, x, y, x+offset, y+height, colour);
	rectfill(bmp, x+offset, y, x+width, y+height, 0);
}

static int getammocolour(AmmoType ammotype)
{
	return ammotype == Hydrofluoric ? makecol(0,192,0) : makecol (192,192,0);
}

void Ship::updatepanel()
{
	blit(panelbackground,panel,0,0,0,0,panel->w,panel->h);
	drawslider(panel, 30, 4, makecol(192,0,0), hydrogen);
	drawslider(panel, 30, 12, makecol(0,0,192), oxygen);
	int ammocolour = getammocolour(ammotype);
	drawslider(panel, 30, 20, ammocolour, ammo);
	drawing_mode(DRAW_MODE_TRANS,0,0,0);
	set_trans_blender(0,0,0,128);
	int shieldcolour = makecol(255,255,255);
	for (float i=0; i<shieldpower; i+= 0.1) {
		int r = (int) (i * 5.0);
		r += 5;
		circle(panel, 11,38, r, shieldcolour);
	}
	drawing_mode(DRAW_MODE_SOLID,0,0,0);
	panelneedsupdate = false;
}

void Ship::draw()
{
	// Draw the connected item line
	if (connecteditem)
	{
		int px = (int) position.x;
		int py = (int) position.y;
		int cx = (int)connecteditem->position.x;
		int cy = (int)connecteditem->position.y;
		line(vscreen, px,
			py,
			cx,
			cy, makecol(255,255,255));
		int minx, miny, maxx, maxy;
		if (px > cx) {
			maxx = px; minx = cx;
		} else {
			maxx = cx; minx = px;
		}
		if (py > cy) {
			maxy = py; miny = cy;
		} else {
			maxy = cy; miny = py;
		}
		addrect(minx,miny, maxx - minx + 1, maxy - miny + 1);
	}

	// Note :: required to differentiate fixed from std::fixed
	::fixed fix_angle = ftofix(angle * (128.0 / M_PI));
	int xoffset = - (sprite->w / 2);
	int yoffset = - (sprite->h / 2);
	rotate_sprite(vscreen, sprite, ((int) position.x) + xoffset, ((int) position.y) + yoffset, fix_angle);
	// circlefill(vscreen, (int) position.x, (int) position.y, 4, makecol(255,255,255));
	// circle(vscreen, (int) position.x, (int) position.y, (int) (radius - 1), makecol(255,255,255));
	// Add dirty rect.
	int x = (int) position.x;
	int y = (int) position.y;
	if (shieldstress >0.25) {
		drawing_mode(DRAW_MODE_TRANS,0,0,0);
		int brightness = (int) (shieldstress * 256);
		if (brightness>255) brightness = 255;
		set_trans_blender(0,0,0,brightness);
		circle(vscreen, (int) position.x, (int) position.y, (int) radius, makecol(255,255,255));
		if (shieldstress>2) {
			brightness = (int) (shieldstress * 64);
			if (brightness>255) brightness = 255;
			set_trans_blender(0,0,0,brightness);
			circle(vscreen, (int) position.x, (int) position.y, ((int) radius) - 1, makecol(255,255,255));
		}
		drawing_mode(DRAW_MODE_SOLID,0,0,0);
	}

	addrect(x -((int) radius) -3 ,y - ((int) radius) -3,
		(int) radius * 2 + 6, (int) radius * 2 + 6);
	if (panelneedsupdate)
		updatepanel();
	drawpanel();
};

void Ship::firerocket(float quantum)
{
	if (criticalhit) return;
	// Check fuel status.
	if ((oxygen <0) || (hydrogen < 0))
		return;
	float sa = sin(angle);
	float ca = cos(angle);
	float impulse = thrust / mass;
	impulse *= PixelsPerMetre;
	float impulsex = (sa * impulse);
	float impulsey = - (ca * impulse);
	velocity.x += impulsex * quantum;
	velocity.y += impulsey * quantum;
	debristime += quantum;
	if (debristime > 0.05) {
		debristime = 0;
		Particle * p = new Particle();
		float randx = randombetween(-10,10);
		float randy = randombetween(-10,10);
		p->position.x = position.x - (sa * radius);
		p->position.y = position.y + (ca * radius);
		p->velocity.x = - (sa * DebrisSpeed) + randx + velocity.x;
		p->velocity.y = ca * DebrisSpeed + randy + velocity.y;

		addobject(p);
	}
	// Consume fuel.
	float fuelused = (quantum * burnrate);
	oxygen -= fuelused;
	hydrogen -= fuelused;
	panelneedsupdate = true;
}

static int collidex, collidey;
static int numcollisions;

static void collidecircleproc(BITMAP *bmp, int x, int y, int d)
{
	if (getpixel(bmp,x,y) != 0) {
		collidex += x;
		collidey += y;
		numcollisions ++;
	}
}

static float meancollisionx, meancollisiony;

static void findmeancollision(int cx, int cy, int r)
{
	collidex = 0;
	collidey = 0;
	numcollisions = 0;
	do_circle(collisionmap, cx, cy, r, 0,
		  collidecircleproc);
	if (numcollisions >0) {
		meancollisionx = (float) collidex;
		meancollisiony = (float) collidey;
		meancollisionx /= (float) numcollisions;
		meancollisiony /= (float) numcollisions;
		meancollisionx = meancollisionx - (float) cx;
		meancollisiony = meancollisiony - (float) cy;
	} else {
		meancollisionx = meancollisiony = 0.0;
	}
}


void CollidesWithScenery::collidedetect()
{
	findmeancollision((int) position.x, (int) position.y, (int) radius);
	if ((fabs(meancollisionx) + fabs(meancollisiony)) > 0.01)
	{
		float strength = sqrt((meancollisionx * meancollisionx) + (meancollisiony * meancollisiony));
		float cofsx = meancollisionx / strength;
		float cofsy = meancollisiony / strength;
		// Work out the component of the velocity in that direction.
		float dotprod = (cofsx * velocity.x) + (cofsy * velocity.y);
		// Apply the appropriate acceleration impulse (delta v)
		if (dotprod > 0) {
			// Collision!
			float factor = -1.5;
			if (dotprod < 3)
			{
				factor = -2.1;
			}
			float accx = factor * (dotprod * cofsx);
			float accy = factor * (dotprod * cofsy);
			velocity.x += accx;
			velocity.y += accy;
			takehit(dotprod);
		}
	}
}

void CollidesWithScenery::tick(float quantum)
{
	// Apply gravity.
	ApplyGravity(this,quantum);
	// Move
	position.x += (velocity.x * quantum);
	position.y += (velocity.y * quantum);
	// Collide detect.
	collidedetect();
}

void CollidesWithScenery::takehit(float strength)
{
	// Nothing
}

void CollidesWithScenery::collidewithother(CollidesWithScenery * other, float strength)
{
	// Nothing.
}

void Ship::collidewithother(CollidesWithScenery * other, float strength)
{
	other->takehit(CollisionDamage); // Made up constant
}

void Ship::blowup()
{
	criticalhit = true;
	criticaltimer = 5;
	shieldstress = 0; // As there are no shields left;
}

void Ship::takehit(float strength)
{
	if (criticalhit) return;
	float damage = strength/40;
	if (shieldstress < damage) {
		// Normal pitch, full vol.
		int vol = (int) damage * 128;
		if (vol > 255) vol = 255;
		// Only play sound if reasonably loud.
		if (vol > 16) {
			play_sound(SHIELD_WAV, 1000, vol);
		}
		shieldstress = damage;
	}
	if (damage > 1.0) {
		shieldpower = shieldpower - (damage / 8);
		panelneedsupdate = true;
		if (shieldpower < 0) {
			blowup();
		}
	}

}

static void makebullet(Ship *ship, float angle)
{
	float sa = sin(angle);
	float ca = cos(angle);
	Shot * s = new Shot(ship);
	s->position.x = ship->position.x + (sa * ship->radius);
	s->position.y = ship->position.y - (ca * ship->radius);
	s->velocity.x = sa * ShotSpeed + ship->velocity.x;
	s->velocity.y = - (ca * ShotSpeed) + ship->velocity.y;
	int ammocolour = getammocolour(ship->ammotype);
	if (ship->ammotype == Hydrofluoric)
		s->damage = s->damage * 3;
	s->colour = ammocolour;
	addobject(s);
}

void Ship::firebullet()
{
	if (firetime >0.01) return;
	if (ammo < 0) return;
	makebullet(this, angle);
	if (playerinfo.weapon == 3) {
		makebullet(this, angle - (M_PI / 12));
		makebullet(this, angle + (M_PI / 12));
	}
	ammo -= ammorate;
	panelneedsupdate = true;
	firetime = FireDelay;
}

void Ship::tick(float quantum)
{
	CollidesWithScenery::tick(quantum);
	float turndir = 0;
	if (! criticalhit)
	{
		if (key[KEY_LEFT]) {
			turndir = -1;
		}
		if (key[KEY_RIGHT]) {
			turndir = 1;
		}
		if (key[KEY_UP]) {
			// fire the rocket.
			firerocket(quantum);
		}
		if (firetime >0)
			firetime -= quantum;
		if (key[KEY_LCONTROL] || key[KEY_RCONTROL]) {
			firebullet();
		}
	} else {
		turndir = 1;
		float moo = randombetween(-20,5);
		if (moo > criticaltimer) {
			Explosion * e = new Explosion();
			e->position.x = position.x + randombetween((int) -radius, (int) radius);
			e->position.y = position.y + randombetween((int) -radius, (int) radius);
			addobject(e);
			Particle * p = new Particle();
			p->position.x = position.x;
			p->position.y = position.y;
			p->velocity.x = randombetween(-30,30);
			p->velocity.y = randombetween(-30,30);
		addobject(p);
		}
		criticaltimer -= quantum;
		if (criticaltimer <0)
		{
			dead = true;
			// half pitch, full vol.
			play_sound(EXPLOSION_WAV, 500, 255);
			for (int j=0; j<40; j++) {
				Particle * p = new Particle();
				p->position.x = position.x;
				p->position.y = position.y;
				p->velocity.x = randombetween(-300,300);
				p->velocity.y = randombetween(-300,300);
				// Extra long life!
				p->life = p->life * 2;
				addobject(p);
			}
		}
	}
	if (playerinfo.engine >= 4)
		turndir = turndir * 2;
	angle += ((turndir * TurnSpeed) * quantum);
	// Normalise the angle.
	if (angle > M_PI)
		angle -= M_PI * 2;
	if (angle < (-M_PI))
		angle += M_PI * 2;
	if (shieldstress > 0) {
		shieldstress -= (quantum * 3);
	}
	// Recharge the shield
	if (shieldpower < 1.0) {
		if ((oxygen > 0) && (hydrogen > 0))
		{
			shieldpower += quantum / ShieldRechargeTime;
			hydrogen -= (quantum / ShieldRechargeTime) * burnrate;
			oxygen -= (quantum / ShieldRechargeTime) * burnrate;
		}
	}
	checkforitems(quantum);
}

void Ship::checkforitems(float quantum)
{
	connecteditem = 0;
	// If you're going too fast, you can't pick stuff up.
	if (((velocity.x * velocity.x) + (velocity.y * velocity.y)) > (MaxPickupSpeed * MaxPickupSpeed))
		return;
	// Check for nearby useful items.
	if (key[KEY_SPACE] && ! criticalhit)
	{
		float itemx = position.x;
		float itemy = position.y + (radius * 3);
		GameObject * obj = findnearest(itemx, itemy);
		StaticObject * so = dynamic_cast<StaticObject *>(obj);
		if (so) {
			if (so->capacity>0) {
				float *targetamount;
				switch (so->type) {
					case TankHydrogen:
						targetamount = &hydrogen;
						break;
					case TankOxygen:
						targetamount = &oxygen;
						break;
					case TankHCl:
						if (ammotype != Hydrochloric) {
							ammo = 0;
							ammotype = Hydrochloric;
						}
						targetamount = &ammo;
						break;
					case TankHF:
						if (ammotype != Hydrofluoric) {
							ammo = 0;
							ammotype = Hydrofluoric;
						}
						targetamount = &ammo;
						break;
					case PlayerBaseExit:
						targetamount = &endlevel;
						break;
					default:
						targetamount = 0;
						break;
				}
				if (targetamount)
				{
					float amount = quantum * TransferRate;
					if (*targetamount < 1.0) {
						so->capacity -= amount;
						*targetamount += amount;
						panelneedsupdate = true;
						connecteditem = so;
						if ((so->capacity<0) && (so->type != PlayerBaseExit)) {
							so->sprite =(BITMAP *) datafile[TANK_EMPTY_TGA].dat;
						}
						if (endlevel > 0.3) {
							playerinfo.levelcomplete = true;
						}
					}
				}
			}
		}
	}
}

void Ship::save(ostream &out)
{
	out << "S " << (int) position.x << " " << (int) position.y << endl;
}

Ship::~Ship()
{
	// Destructor
	destroy_bitmap(panel);
};

// PARTICLE

Particle::Particle() // Default constructor.
{
	life = 2.5; // seconds.
	colour = makecol(255,255,255);
	dead = false;
}

void Particle::draw()
{
	// Draw a point
	int ix = (int) position.x;
	int iy = (int) position.y;
	int trans = 255;
	if (life < 2) {
		trans = (int) (life * 128);
		if (trans > 255) trans = 255;
	}
	drawing_mode(DRAW_MODE_TRANS,0,0,0);
	set_trans_blender(0,0,0,trans);
	putpixel(vscreen, ix, iy, colour);
	set_trans_blender(0,0,0,trans / 2);
	putpixel(vscreen, ix+1, iy, colour);
	putpixel(vscreen, ix, iy+1, colour);
	putpixel(vscreen, ix-1, iy, colour);
	putpixel(vscreen, ix, iy-1, colour);
	drawing_mode(DRAW_MODE_SOLID,0,0,0);
	addrect(ix-1, iy-1, 3, 3);
}

void Particle::tick(float quantum)
{
	life -= quantum;
	// Move
	ApplyGravity(this, quantum);
	position.x += (velocity.x * quantum);
	position.y += (velocity.y * quantum);
	int ix = (int) position.x;
	int iy = (int) position.y;
	if (life <0)
		dead = true;
	if (getpixel(collisionmap,ix - 2, iy) != 0) {
		velocity.x = fabs(velocity.x) * 0.6;
	}
	if (getpixel(collisionmap,ix + 2, iy) != 0) {
		velocity.x = - fabs(velocity.x) * 0.6;
	}
	if (getpixel(collisionmap,ix, iy - 2) != 0) {
		velocity.y = fabs(velocity.y) * 0.6;
	}
	if (getpixel(collisionmap,ix, iy + 2) != 0) {
		velocity.y = - fabs(velocity.y) * 0.6;
	}
}

Particle::~Particle()
{

}

/*** STATICOBJECT **/

StaticObject::StaticObject(StaticType s)
{
	type = s;
	int bmid = -1;
	switch (type) {
		case PlayerBaseExit:
			bmid = BASE_TGA;
			break;
		case PlayerBaseStart:
			bmid = BASE_START_TGA;
			break;
		case TankHydrogen:
			bmid = TANK_HYDROGEN_TGA;
			break;
		case TankOxygen:
			bmid = TANK_OXYGEN_TGA;
			break;
		case TankHCl:
			bmid = TANK_HCL_TGA;
			break;
		case TankHF:
			bmid = TANK_HF_TGA;
			break;
	}
	if (bmid == -1)
		throw "Unable to get bitmap in StaticObject";
	sprite = (BITMAP *) datafile[bmid].dat;
	drawn = false;
	width = sprite->w;
	height = sprite->h;
	dead = false;
	capacity = 1.0; // How much it contains (if a tank)
}

void StaticObject::draw()
{
	int px, py;
	px = ((int) position.x) - (width / 2);
	py = ((int) position.y) - height;
	draw_sprite(vscreen, sprite, px, py);
	// Only a need to update the rect if something has changed.
	if (! drawn) {
		addrect(px,py, width, height);
		drawn = true;
	}
}

static void spritetocollisionmap(BITMAP *sprite, int px, int py, bool add, int id)
{
	int colour = add ? 8+id : 0;
	int maskColour = bitmap_mask_color(sprite);
	for (int x=0; x< sprite->w; x++)
		for (int y = 0; y < sprite->h; y++) {
		if (getpixel(sprite,x,y) != maskColour) {
			putpixel(collisionmap, px+x, py+y, colour);
		}
	}
}

void StaticObject::drawtocollisionmap()
{
	int px, py;
	px = ((int) position.x) - (width / 2);
	py = ((int) position.y) - height;
	spritetocollisionmap(sprite, px, py, true, (int) type);
}

void StaticObject::save(ostream &out)
{
	int typenum = (int) type;
	out << "T " << typenum << " " << (int) position.x << " " << (int) position.y << endl;
}

StaticObject::~StaticObject()
{
	// Nothing
}

/** EXPLOSION **/
Explosion::Explosion()
{
	dead = false;
	life = ((float) randombetween(15,80)) / 100;
	velocity.x = 0;
	velocity.y = 0;
	sprite = (BITMAP *) datafile[EXPLOSION_TGA].dat;
}

void Explosion::draw()
{
	int px, py;
	px = ((int) position.x) - (sprite->w / 2);
	py = ((int) position.y) - (sprite->h / 2);
	int orient = randombetween(0,3);
	switch (orient) {
		case 0:
			draw_sprite(vscreen,sprite,px,py);
			break;
		case 1:
			draw_sprite_h_flip(vscreen,sprite,px,py);
			break;
		case 2:
			draw_sprite_v_flip(vscreen,sprite,px,py);
			break;
		case 3:
			draw_sprite_vh_flip(vscreen,sprite,px,py);
			break;
	}
	addrect(px,py,sprite->w,sprite->h);
}
Explosion::~Explosion()
{
	// Nothing
}

