#include <vector>
#include <allegro.h>
#include <math.h>
#include <iostream>

using namespace std;

#include "common.h"
#include "classes.h"
#include "data.h"
#include "manager.h"
#include "shot.h"

const float ShotDamage = 100;

/*
 * Shot used for player and enemy.
 */

 Shot::Shot(GameObject * newowner)
 {
 	dead = false;
	owner = newowner;
	hp = 1.0;
	radius = 3.0;
	colour = makecol(255,255,255);
	life = 6;
	mass = 1000.0; // Kilograms
	damage = ShotDamage;
	// Double pitch, half vol.
	play_sound(BANG_WAV, 2000, 128);
}

void Shot::draw()
{
	int px = (int) position.x;
	int py = (int) position.y;
	int radint = (int) radius;
	drawing_mode(DRAW_MODE_TRANS,0,0,0);
	set_trans_blender(0,0,0,128);
	circlefill(vscreen,px,py, radint, colour);
	drawing_mode(DRAW_MODE_SOLID,0,0,0);
	circlefill(vscreen,px,py, (int) radius/2, colour);
	// Add more to radint, so we get the rect big enough.
	radint++;
	addrect(px - radint, py - radint, radint * 2, radint * 2);
}

void Shot::tick(float quantum)
{
	CollidesWithScenery::tick(quantum);
	life -= quantum;
	if (life <0)
		dead = true;
}

// No save method.

void Shot::takehit(float strength)
{
	hp = hp - (strength / 40);
	if (hp < 0) {
		dead = true;
		// Normal pitch, full vol.
		play_sound(BANG_WAV, 1000, 255);
		// Make some particles.
		for (int i=0; i<5; i++) {
			Particle * p = new Particle();
			float randx = randombetween(-40,40);
			float randy = randombetween(-40,40);
			p->position.x = position.x;
			p->position.y = position.y;
			p->velocity.x = randx;
			p->velocity.y = randy;
			p->colour = colour;
			addobject(p);
		}
		Explosion * e = new Explosion();
		e->position.x = position.x;
		e->position.y = position.y;
		addobject(e);
	}
}

void Shot::collidewithother(CollidesWithScenery * other, float strength)
{
	other->takehit(damage); // Made up constant
}

Shot::~Shot()
{
	// nothing
}
